package map.map25000;

import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import util.Log;

import map.data.Curve;
import map.data.CoastNode;
import map.data.Facility;
import map.data.City;
import map.data.Label;
import map.data.Mesh;
import map.data.Node;
import map.data.Road;
import map.data.Station;

/**
 * 数値地図25000のデータファクトリー
 * 
 * @author Masayasu Fujiwara
 * 
 */
public class Map25000Factory extends Thread {

	private final int DEFAULT_SLP_CAPACITY = 20000;

	private final int DEFAULT_MESH_CAPACITY = 20000;

	private final int DEFAULT_NODE_CAPACITY = 1000;

	private final int DEFAULT_BORDER_CAPACITY = 20;

	/**
	 * ファイル読み込みのための正規表現
	 */
	private final Pattern SAL_PARSER = Pattern.compile("([A-Z]{2})" + "(?:"
			+ "(?:" + "\\(" + "([^)]*)" + "\\)" + ")" + "|" + "\\{"
			+ "([^\\}\\{]*)" + "\\}" + ")" + "(?:" + "\\{" + "(?:" + "("
			+ "(?:" + "\\d{6},)*" + "\\d{6}" + ")" + "|" + "(" + "(?:"
			+ "[A-Z]{2}" + "\\(" + "[^\\)]*" + "\\)" + ")*" + ")" + ")" + "\\}"
			+ ")?");

	/**
	 * 必要なファイルを取得するためのクラス
	 */
	private final Map25000Storage storage;

	private final String KEY_RAIL = "rail";

	private final String KEY_STATION = "station";

	private final String KEY_NODE = "node";

	private final String KEY_ROAD = "road";

	private final String KEY_MESH = "mesh";

	private final String KEY_SLP = "slp";

	private final String KEY_RIVER = "river";

	private final String KEY_BORDER = "border";

	private String KEY_COAST;

	/**
	 * コンストラクタ
	 * 
	 * @param storage
	 *            数値地図25000のデータ取得管理クラス
	 */
	public Map25000Factory(Map25000Storage storage) {
		this.storage = storage;
	}

	/**
	 * 展開
	 * 
	 * @param nodes
	 * @param set
	 */
	private void extract(Collection<CoastNode> nodes, Collection<Curve> set) {
		Set<Curve> borderSet = new HashSet<Curve>();
		for (CoastNode node : nodes) {
			if (node.getBorder().size() == 2) {
				for (int i = 0; i < 4; i++) {
					Curve border = node.connect(i);
					if (border != null) {
						set.add(border);
					}
				}
			}
		}
		for (CoastNode node : nodes) {
			for (Curve border : node.getBorder()) {
				borderSet.add(border);
			}
		}
		for (Curve border : borderSet) {
			set.add(border);
		}
	}

	/**
	 * 海岸線の展開
	 * 
	 * @param nodes
	 * @param map
	 */
	private void extractCoast(Collection<CoastNode> nodes,
			Map<Curve, Boolean> map) {
		Set<Curve> borderSet = new HashSet<Curve>();
		for (CoastNode node : nodes) {
			if (node.getBorder().size() == 2) {
				for (int i = 0; i < 4; i++) {
					Curve border = node.connect(i);
					if (border != null) {
						map.put(border, true);
					}
				}
			}
		}
		for (CoastNode node : nodes) {
			for (Curve border : node.getBorder()) {
				borderSet.add(border);
			}
		}
		for (Curve border : borderSet) {
			map.put(border, false);
		}
	}

	/**
	 * 市区町村番号からファイルを取得
	 * 
	 * @param code
	 *            市区町村番号
	 * @param fileMap
	 * @return slmファイル
	 * @throws IOException
	 *             入出力エラー
	 */
	private File getFile(int code, Map<String, File> fileMap)
			throws IOException {
		File slm = null;
		for (File file : this.storage.getStreaming(code)) {
			Log.out(this, "get "+  file.getCanonicalPath());
			String path = file.getPath();
			if (path.endsWith("slm")) {
				slm = file;
				if (fileMap == null) {
					break;
				}
			} else if (fileMap != null) {
				if (path.endsWith(".sal")) {
					if (path.endsWith("DS.sal")) {
						// 道路接点
						fileMap.put("DS.sal", file);
					} else if (path.endsWith("DK.sal")) {
						// 道路区間
						fileMap.put("DK.sal", file);
					} else if (path.endsWith("SK.sal")) {
						// 水域界
						fileMap.put("SK.sal", file);
					} else if (path.endsWith("GK.sal")) {
						// 行政界
						fileMap.put("GK.sal", file);
					} else if (path.endsWith("CM.sal")) {
						// 市町村名
						fileMap.put("CM.sal", file);
					} else if (path.endsWith("MH.sal")) {
						// メッシュ標高
						fileMap.put("MH.sal", file);
					} else if (path.endsWith("KK.sal")) {
						// 河川区間
						fileMap.put("KK.sal", file);
					} else if (path.endsWith("KO.sal")) {
						// 施設
						fileMap.put("KO.sal", file);
					} else if (path.endsWith("TK.sal")) {
						// 鉄道区間
						fileMap.put("TK.sal", file);
					} else if (path.endsWith("EK.sal")) {
						// 駅区間
						fileMap.put("EK.sal", file);
					}
				} else if (path.endsWith(".slp")) {
					if (path.endsWith("MH.slp")) {
						// メッシュ標高座標
						fileMap.put("MH.slp", file);
					} else {
						// 座標
						fileMap.put(".slp", file);
					}
				}
			}
		}
		return slm;
	}

	/**
	 * 多角形面の向き
	 * 
	 * @param x
	 *            多角形面を構成するx座標の配列
	 * @param y
	 *            多角形面を構成するy座標の配列
	 * @return 多角形から計算した外積のZ座標の値
	 */
	public int polygonDirection(int[] x, int[] y) {
		int vector = 0;
		for (int i = 1; i < x.length - 1; i++) {
			int dx0 = x[i] - x[0];
			int dy0 = y[i] - y[0];

			int dx1 = x[i + 1] - x[0];
			int dy1 = y[i + 1] - y[0];

			vector += dx1 * dy0 - dy1 * dx0;
		}
		return (vector >= 0) ? 1 : -1;
	}

	/**
	 * 市区町村の範囲を取得します。
	 * 
	 * @param city
	 *            市区町村データ
	 * @throws IOException
	 *             入出力エラー
	 */
	public void preproduct(City city) throws IOException {
		Log.out(this, "preproduct started.");
		Rectangle area = this.readSLM(city.getCode(), null);
		city.setArea(area);
		Log.out(this, "preproduct finished.");
	}

	/**
	 * 市区町村データを取得
	 * 
	 * @param code
	 *            市区町村番号
	 * @return 市区町村データ
	 * @throws IOException
	 *             入出力エラー
	 */
	public City preproduct(int code) throws IOException {
		Rectangle area = this.readSLM(code, null);
		return new City(code, area);
	}

	/**
	 * 市区町村データを生成
	 * 
	 * @param city
	 *            市区町村データ
	 * @throws IOException
	 *             入出力エラー
	 */
	public void product(City city) throws IOException {

		Map<String, File> fileMap = new HashMap<String, File>();

		int code = city.getCode();
		this.getFile(code, fileMap);

		Mesh[] mesh = null;
		if (this.storage.hasSerializable(code, this.KEY_MESH)) {
			mesh = (Mesh[]) this.storage.readSerializable(code, "mesh");
		}
		if (mesh == null) {
			mesh = this.readMesh(fileMap.get("MH.slp"), fileMap.get("MH.sal"),
					city.getArea()).toArray(new Mesh[] {});
			this.storage.writeSerializable(code, "mesh", mesh);
		}

		// ファイルの読み込み
		Point[] slp = null;
		if (this.storage.hasSerializable(code, this.KEY_SLP)) {
			slp = (Point[]) this.storage.readSerializable(code, this.KEY_SLP);
		}
		if (slp == null) {
			slp = this.readSLP(fileMap.get(".slp"), city.getArea());
			this.storage.writeSerializable(code, this.KEY_SLP, slp);
		}

		Map<String, Facility[]> label = new HashMap<String, Facility[]>();
		label.put(Label.KEY_CM, this.readFacility(fileMap.get("CM.sal"), slp));
		label.put(Label.KEY_KO, this.readFacility(fileMap.get("KO.sal"), slp));

		// 道路接点の読み込み
		Map<Integer, Node> nodes = null;
		if (this.storage.hasSerializable(code, this.KEY_NODE)
				&& this.storage.hasSerializable(city.getCode(), this.KEY_ROAD)) {
			nodes = (Map<Integer, Node>) this.storage.readSerializable(code, this.KEY_NODE);
		}
		if (nodes == null) {
			nodes = this.readNode(code, fileMap.get("DS.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_NODE, nodes);
		}

		// 道路区間の読み込み
		Road[][] road = this.readRoad(code, fileMap.get("DK.sal"), slp, nodes,
				true);

		Curve[][] rail = null;
		Station[] station = null;
		if (this.storage.hasSerializable(code, this.KEY_RAIL)
				&& this.storage.hasSerializable(city.getCode(),
						this.KEY_STATION)) {
			rail = (Curve[][]) this.storage.readSerializable(code,
					this.KEY_RAIL);
			station = (Station[]) this.storage.readSerializable(code,
					this.KEY_STATION);
		}
		if (rail == null || station == null) {
			Map<Integer, Curve> railMap = new HashMap<Integer, Curve>();
			// 鉄道区間の読み込み
			rail = this.readRail(fileMap.get("TK.sal"), slp, railMap);
			// 駅区間の読み込み
			station = this.readStation(fileMap.get("EK.sal"), railMap);
			this.storage.writeSerializable(code, this.KEY_RAIL, rail);
			this.storage.writeSerializable(code, this.KEY_STATION, station);
		}

		// 河川区間の読み込み
		Curve[] river = null;
		if (this.storage.hasSerializable(code, this.KEY_RIVER)) {
			river = (Curve[]) this.storage.readSerializable(code,
					this.KEY_RIVER);
		}
		if (river == null) {
			river = this.readBorder(fileMap.get("KK.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_RIVER, river);
		}

		// 水域界の読み込み
		Map<Curve, Boolean> coast = null;
		if (this.storage.hasSerializable(code, this.KEY_COAST)) {
			coast = (Map<Curve, Boolean>) this.storage.readSerializable(code,
					this.KEY_COAST);
		}
		if (river == null) {
			coast = this.readCoast(fileMap.get("SK.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_COAST, coast);
		}

		// 行政界の読み込み
		Curve[] border = null;
		if (this.storage.hasSerializable(code, this.KEY_BORDER)) {
			border = (Curve[]) this.storage.readSerializable(code,
					this.KEY_BORDER);
		}
		if (river == null) {
			border = this.readBorder(fileMap.get("GK.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_BORDER, border);
		}

		city.setData(border, coast, river, rail, station, nodes, road, label,
				mesh);
	}

	/**
	 * 市区町村データを生成
	 * 
	 * @param code
	 *            市区町村番号
	 * @return 市区町村データ
	 * @throws IOException
	 */
	public City product(int code) throws IOException {
		Map<String, File> fileMap = new HashMap<String, File>();
		Rectangle area = this.readSLM(code, fileMap);

		Mesh[] mesh = null;
		if (this.storage.hasSerializable(code, this.KEY_MESH)) {
			mesh = (Mesh[]) this.storage.readSerializable(code, this.KEY_MESH);
		}
		if (mesh == null) {
			mesh = this.readMesh(fileMap.get("MH.slp"), fileMap.get("MH.sal"),
					area).toArray(new Mesh[] {});
			this.storage.writeSerializable(code, this.KEY_MESH, mesh);
		}

		// ファイルの読み込み
		Point[] slp = null;
		if (this.storage.hasSerializable(code, this.KEY_SLP)) {
			slp = (Point[]) this.storage.readSerializable(code, this.KEY_SLP);
		}
		if (slp == null) {
			slp = this.readSLP(fileMap.get(".slp"), area);
			this.storage.writeSerializable(code, this.KEY_SLP, slp);
		}

		Map<String, Facility[]> label = new HashMap<String, Facility[]>();
		label.put(Label.KEY_CM, this.readFacility(fileMap.get("CM.sal"), slp));
		label.put(Label.KEY_KO, this.readFacility(fileMap.get("KO.sal"), slp));

		// 道路接点の読み込み
		Map<Integer, Node> nodes = null;
		if (this.storage.hasSerializable(code, this.KEY_NODE)
				&& this.storage.hasSerializable(code, this.KEY_ROAD)) {
			nodes = (Map<Integer, Node>) this.storage.readSerializable(code,
					this.KEY_NODE);
		}
		if (nodes == null) {
			nodes = this.readNode(code, fileMap.get("DS.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_NODE, nodes);
		}
		// 道路区間の読み込み
		Road[][] road = this.readRoad(code, fileMap.get("DK.sal"), slp, nodes,
				true);

		Curve[][] rail = null;
		Station[] station = null;
		if (this.storage.hasSerializable(code, this.KEY_RAIL)
				&& this.storage.hasSerializable(code, this.KEY_STATION)) {
			rail = (Curve[][]) this.storage.readSerializable(code,
					this.KEY_RAIL);
			station = (Station[]) this.storage.readSerializable(code,
					this.KEY_STATION);
		}
		if (rail == null || station == null) {
			Map<Integer, Curve> railMap = new HashMap<Integer, Curve>();
			// 鉄道区間の読み込み
			rail = this.readRail(fileMap.get("TK.sal"), slp, railMap);
			// 駅区間の読み込み
			station = this.readStation(fileMap.get("EK.sal"), railMap);
			this.storage.writeSerializable(code, this.KEY_RAIL, rail);
			this.storage.writeSerializable(code, this.KEY_STATION, station);
		}

		// 河川区間の読み込み
		Curve[] river = null;
		if (this.storage.hasSerializable(code, this.KEY_RIVER)) {
			river = (Curve[]) this.storage.readSerializable(code,
					this.KEY_RIVER);
		}
		if (river == null) {
			river = this.readBorder(fileMap.get("KK.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_RIVER, river);
		}

		// 水域界の読み込み
		Map<Curve, Boolean> coast = null;
		if (this.storage.hasSerializable(code, this.KEY_COAST)) {
			coast = (Map<Curve, Boolean>) this.storage.readSerializable(code, this.KEY_COAST);
		}
		if (river == null) {
			coast = this.readCoast(fileMap.get("SK.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_COAST, coast);
		}

		// 行政界の読み込み
		Curve[] border = null;
		if (this.storage.hasSerializable(code, this.KEY_BORDER)) {
			border = (Curve[]) this.storage.readSerializable(code,
					this.KEY_BORDER);
		}
		if (river == null) {
			border = this.readBorder(fileMap.get("GK.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_BORDER, border);
		}

		// マップデータを返す
		return new City(code, area, border, coast, river, rail, station, nodes,
				road, label, mesh);
	}

	/**
	 * 市区町村データに頂点データを付加します。
	 * 
	 * @param city
	 *            市区町村データ
	 * @param isDetailRoad
	 *            詳細な道路データを作成する場合はtrue
	 * @throws IOException
	 *             入出力エラー
	 */
	public void productNode(City city, boolean isDetailRoad) throws IOException {

		Map<String, File> fileMap = new HashMap<String, File>();

		int code = city.getCode();

		if (city.hasData()) {
			this.getFile(code, fileMap);
		} else {
			city.setArea(this.readSLM(code, fileMap));
		}

		// ファイルの読み込み
		Point[] slp = null;
		if (this.storage.hasSerializable(code, this.KEY_SLP)) {
			slp = (Point[]) this.storage.readSerializable(code,
					this.KEY_SLP);
		}
		if (slp == null) {
			slp = this.readSLP(fileMap.get(".slp"), city.getArea());
			this.storage.writeSerializable(code, this.KEY_SLP, slp);
		}

		// 道路接点の読み込み
		Map<Integer, Node> nodes = null;
		if (this.storage.hasSerializable(code, this.KEY_NODE)) {
			nodes = (Map<Integer, Node>) this.storage.readSerializable(code, this.KEY_NODE);
			this.storage
					.writeSerializable(code, this.KEY_NODE, nodes);
		}
		if (nodes == null) {
			nodes = this.readNode(code, fileMap.get("DS.sal"), slp);
		}
		this.readRoad(code, fileMap.get("DK.sal"), slp, nodes,
				isDetailRoad);
		city.setData(nodes);
	}

	/**
	 * 頂点データを生成します。
	 * 
	 * @param code
	 *            市区町村番号
	 * @param isDetailRoad
	 *            詳細な道路データを作成する場合はtrue
	 * @return 頂点データ
	 * @throws IOException
	 *             入出力エラー
	 */
	public City productNode(int code, boolean isDetailRoad) throws IOException {
		Map<String, File> fileMap = new HashMap<String, File>();
		Rectangle area = this.readSLM(code, fileMap);
		City map = new City(code, area);

		// ファイルの読み込み
		Point[] slp = null;
		if (this.storage.hasSerializable(code, this.KEY_SLP)) {
			slp = (Point[]) this.storage.readSerializable(code, this.KEY_SLP);
		}
		if (slp == null) {
			slp = this.readSLP(fileMap.get(".slp"), area);
			this.storage.writeSerializable(code, this.KEY_SLP, slp);
		}

		// 道路接点の読み込み
		Map<Integer, Node> nodes = null;
		if (this.storage.hasSerializable(code, this.KEY_NODE)) {
			nodes = (Map<Integer, Node>) this.storage.readSerializable(code,
					this.KEY_NODE);
		}
		if (nodes == null) {
			nodes = this.readNode(code, fileMap.get("DS.sal"), slp);
			this.storage.writeSerializable(code, this.KEY_NODE, nodes);
		}
		this.readRoad(code, fileMap.get("DK.sal"), slp, nodes, isDetailRoad);

		map.setData(nodes);
		return map;
	}

	/**
	 * 境界の読み込み
	 * 
	 * @param file
	 * @param slp
	 * @return 境界の配列
	 * @throws IOException
	 */
	private Curve[] readBorder(File file, Point[] slp) throws IOException {
		List<Curve> border = new ArrayList<Curve>(this.DEFAULT_BORDER_CAPACITY);
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				int[] curveX = null;
				int[] curveY = null;
				while (matcher.find()) {
					String id = matcher.group(1);
					if (id.equals("CV")) {
						// カーブメモリ
						String[] ref = matcher.group(4).split(",");
						curveX = new int[ref.length];
						curveY = new int[ref.length];
						for (int i = 0; i < ref.length; i++) {
							Point point = slp[Integer.parseInt(ref[i]) - 1];
							curveX[i] = point.x;
							curveY[i] = point.y;
						}
					}
				}
				Curve b = new Curve(curveX, curveY, 0);
				border.add(b);
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return border.toArray(new Curve[] {});
	}

	/**
	 * 水域界の読み込み
	 * 
	 * @param file
	 *            水域界のファイル
	 * @param slp
	 * @return 水域界データ
	 * @throws IOException
	 */
	private Map<Curve, Boolean> readCoast(File file, Point[] slp)
			throws IOException {
		Map<Curve, Boolean> map = new HashMap<Curve, Boolean>();
		Map<Point, CoastNode> points = new HashMap<Point, CoastNode>();
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				int type = 0;
				int[] curveX = null;
				int[] curveY = null;
				while (matcher.find()) {
					String id = matcher.group(1);
					if (id.equals("CV")) {
						// カーブメモリ
						String[] ref = matcher.group(4).split(",");
						curveX = new int[ref.length];
						curveY = new int[ref.length];
						for (int i = 0; i < ref.length; i++) {
							Point point = slp[Integer.parseInt(ref[i]) - 1];
							curveX[i] = point.x;
							curveY[i] = point.y;
						}
					} else if (id.equals("SR")) {
						/**
						 * 51 水涯線または湖岸線 52 海岸線 53 河口 54 湖沼と河川の境界
						 */
						type = Integer.parseInt(matcher.group(3)) - 51;
						if (type != 1) {
							type = 0;
						}
						if (type == 3) {
							Log.out(this, "SR 54 ? " + file);
						}
					}
				}
				Curve border = new Curve(curveX, curveY, type);
				CoastNode p1 = new CoastNode(curveX[0], curveY[0]);
				CoastNode p2 = new CoastNode(curveX[curveX.length - 1],
						curveY[curveX.length - 1]);
				if (p1.equals(p2)) {
					map.put(border, true);
				} else if (border.getType() == 1) {
					map.put(border, false);
				} else {
					if (points.containsKey(p1)) {
						p1 = points.get(p1);
					} else {
						points.put(p1, p1);
					}
					if (points.containsKey(p2)) {
						p2 = points.get(p2);
					} else {
						points.put(p2, p2);
					}
					Curve b = p1.put(p2, border);
					if (b != null) {
						map.put(b, true);
					}
				}
			}
			this.extractCoast(points.values(), map);
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return map;
	}

	/**
	 * 施設情報の読み込み
	 * 
	 * @param file
	 * @param slp
	 * @throws IOException
	 * @return 施設情報配列
	 */
	private Facility[] readFacility(File file, Point[] slp) throws IOException {
		List<Facility> facility = new ArrayList<Facility>();
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				Point point = null;
				boolean isNM = false;
				String name = null;
				while (matcher.find()) {
					String param = matcher.group(1);
					if (param.equals("NM")) {
						name = matcher.group(3);
						isNM = true;
					} else if (param.equals("PT")) {
						point = slp[Integer.parseInt(matcher.group(4)) - 1];
					}
					if (isNM && point != null) {
						facility.add(new Facility(name, point.x, point.y));
					}
				}
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return facility.toArray(new Facility[] {});
	}

	/**
	 * 
	 * メッシュ標高の読み込み
	 * 
	 * @param slp
	 * @param sal
	 * @param area
	 * @return メッシュ標高のリスト
	 * @throws IOException
	 * 
	 */
	private List<Mesh> readMesh(File slp, File sal, Rectangle area)
			throws IOException {
		List<Point> points = new ArrayList<Point>(this.DEFAULT_MESH_CAPACITY);
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new FileReader(slp));
			while (bi.ready()) {
				String line = bi.readLine();
				int x = Integer.parseInt(line.substring(0, 7)) + area.x;
				int y = Integer.parseInt(line.substring(8, 15)) + area.y;
				points.add(new Point(x, y));
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		List<Mesh> mesh = new ArrayList<Mesh>(points.size());
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					sal), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				// String category = line.substring(0, 2);
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				Point point = null;
				int height = 0;
				while (matcher.find()) {
					String param = matcher.group(1);
					if (param.equals("PT")) {
						point = points
								.get(Integer.parseInt(matcher.group(4)) - 1);
					} else if (param.equals("HK")) {
						height = Integer.parseInt(matcher.group(3));
					}
				}
				mesh.add(new Mesh(point.x, point.y, height));
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return mesh;
	}

	/**
	 * 接点の読み込み
	 * 
	 * @param code
	 *            市区町村番号
	 * @param file
	 * @param slp
	 * @return 頂点番号をkey、頂点をvalueとしたMap
	 * @throws IOException
	 *             入出力エラー
	 */
	private Map<Integer, Node> readNode(int code, File file, Point[] slp)
			throws IOException {
		Map<Integer, Node> node = new HashMap<Integer, Node>(
				this.DEFAULT_NODE_CAPACITY);
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				String id = null;
				Point point = null;
				int slpindex = -1;
				while (matcher.find()) {
					String param = matcher.group(1);
					if (param.equals("PT")) {
						slpindex = Integer.parseInt(matcher.group(4)) - 1;
						point = slp[slpindex];
					} else if (param.equals("ND")) {
						id = matcher.group(2).substring(5, 11);
					}
				}
				if (point != null && id != null) {
					if (code == 40101) {
						if (id.equals("000985")) {
							point.setLocation(471453115, 122261236);
						} else if (id.equals("001020")) {
							point.setLocation(471457816, 122265788);
						}
					} else if (code == 38202 && id.equals("001867")) {
						point.setLocation(478795248, 122834185);
					}
					node.put(Integer.parseInt(id), new Node(Long.parseLong(code
							+ id), point));
				}
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return node;
	}

	/**
	 * 水域界の読み込み
	 * 
	 * @param file
	 *            水域界のファイル
	 * @param slp
	 * @param map
	 * @return 水域界データ
	 * @throws IOException
	 */
	private Curve[][] readRail(File file, Point[] slp, Map<Integer, Curve> map)
			throws IOException {
		Collection<Curve> jr = new HashSet<Curve>();
		Collection<Curve> other = new HashSet<Curve>();
		Map<Point, CoastNode> points = new HashMap<Point, CoastNode>();
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				int type = 0;
				int[] curveX = null;
				int[] curveY = null;
				while (matcher.find()) {
					String id = matcher.group(1);
					if (id.equals("CV")) {
						// カーブメモリ
						String[] ref = matcher.group(4).split(",");
						curveX = new int[ref.length];
						curveY = new int[ref.length];
						for (int i = 0; i < ref.length; i++) {
							Point point = slp[Integer.parseInt(ref[i]) - 1];
							curveX[i] = point.x;
							curveY[i] = point.y;
						}
					} else if (id.equals("SB")) {
						type = Integer.parseInt(matcher.group(3)) - 43;
					}
				}
				Curve border = new Curve(curveX, curveY, type);
				map.put(Integer.valueOf(line.substring(8, 14)), border);
				CoastNode p1 = new CoastNode(curveX[0], curveY[0]);
				CoastNode p2 = new CoastNode(curveX[curveX.length - 1],
						curveY[curveX.length - 1]);
				if (border.getType() != 0) {
					other.add(border);
				} else if (p1.equals(p2)) {
					jr.add(border);
				} else {
					if (points.containsKey(p1)) {
						p1 = points.get(p1);
					} else {
						points.put(p1, p1);
					}
					if (points.containsKey(p2)) {
						p2 = points.get(p2);
					} else {
						points.put(p2, p2);
					}
					Curve b = p1.put(p2, border);
					if (b != null) {
						jr.add(b);
					}
				}
			}
			this.extract(points.values(), jr);
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return new Curve[][] { jr.toArray(new Curve[] {}),
				other.toArray(new Curve[] {}) };
	}

	/**
	 * 道路の読み込み
	 * 
	 * @param code
	 * @param file
	 *            道路情報のファイル
	 * @param slp
	 *            slpファイルの読み込んだもの
	 * @param node
	 * @param isDetail
	 * @return 道路ファイル
	 * @throws IOException
	 */
	private Road[][] readRoad(int code, File file, Point[] slp,
			Map<Integer, Node> node, boolean isDetail) throws IOException {
		List<List<Curve>> border = new ArrayList<List<Curve>>(6);
		List<List<Curve>> highway = new ArrayList<List<Curve>>(6);
		if (isDetail) {
			for (int i = 0; i < 6; i++) {
				border.add(new ArrayList<Curve>());
				highway.add(new ArrayList<Curve>());
			}
		}
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				int[] curveX = null;
				int[] curveY = null;
				int type = 0;
				int width = 0;
				int n0 = -1;
				int n1 = -1;
				String[] ref = null;
				double cost = 0;
				while (matcher.find()) {
					String id = matcher.group(1);
					if (id.equals("CV")) {
						// カーブメモリ
						ref = matcher.group(4).split(",");
						curveX = new int[ref.length];
						curveY = new int[ref.length];
						Point p0 = null;
						for (int i = 0; i < ref.length; i++) {
							Point point = slp[Integer.parseInt(ref[i]) - 1];
							curveX[i] = point.x;
							curveY[i] = point.y;
							if (i == 0) {
								p0 = point;
							} else {
								long dx = p0.x - point.x;
								long dy = p0.y - point.y;
								cost += Math.sqrt(dx * dx + dy * dy);
								p0 = point;
							}
						}
					} else if (id.equals("EG")) {
						String edge = matcher.group(5);
						n0 = Integer.parseInt(edge.substring(13, 19));
						n1 = Integer.parseInt(edge.substring(34, 40));
					} else if (id.equals("SB")) {
						type = Integer.parseInt(matcher.group(3)) - 13;
						if (type < 0) {
							type = 0;
						}
					} else if (id.equals("FI")) {
						// 17から16進数で与えられる
						width = Integer.parseInt(matcher.group(3), 16) - 23;
						if (width == 7) {
							width = 5;
						}
					}
				}
				Road b;
				if (isDetail) {
					b = new Road(curveX, curveY, type, width, (float) cost);
				} else {
					b = new Road(type, width, (float) cost);
				}
				Node node0 = node.get(n0);
				Node node1 = node.get(n1);
				node0.connect(code * 1000000L + n1, b);
				node1.connect(code * 1000000L + n0, b);
				if (isDetail) {
					if (type == 3) {
						highway.get(width).add(b);
					} else {
						border.get(width).add(b);
					}
				}
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		InputStream stream = this.storage.getBoundaryNode(code);
		if (stream != null) {
			DataInputStream disc = null;
			try {
				disc = new DataInputStream(new BufferedInputStream(stream));
				while (disc.available() >= 12) {
					node.get(disc.readInt()).connect(disc.readLong(), null);
				}
			} finally {
				if (disc != null) {
					disc.close();
				}
			}
		}
		Road[][] road = new Road[12][];
		if (isDetail) {
			for (int i = 0; i < 6; i++) {
				road[i] = border.get(i).toArray(new Road[] {});
			}
			for (int i = 0; i < 6; i++) {
				road[i + 6] = highway.get(i).toArray(new Road[] {});
			}
			return road;
		} else {
			return null;
		}
	}

	/**
	 * 領域などの読み込み
	 * 
	 * @param code
	 *            市区町村番号
	 * @param map
	 * @return 市区町村の領域
	 * @throws IOException
	 */
	private Rectangle readSLM(int code, Map<String, File> map) throws IOException {
		Rectangle rect = null;
		BufferedReader bi = null;
		if (map != null || this.storage.isAvailable(City.cityFormat(code)) || code == 13421) {
			if (code == 13421) {
				Log.out(this, "getStreaming start");
				this.storage.getStreaming(code);
				Log.out(this, "getStreaming finish");
				int[] codes = new int[] { 13425, 13426, 13427, 13428, 13429 };
				rect = new Rectangle();
				for (int i = 0; i < codes.length; i++) {
					Log.out(this, "read slm " + codes[i]);
					rect.add(this.readSLM(codes[i], map));
				}
			} else {
				Log.out(this, "read slm "+ code);
				File file = this.getFile(code, map);
				try {
					bi = new BufferedReader(new FileReader(file));
					// 原点読み込み
					StringTokenizer st = new StringTokenizer(bi.readLine(), ",");
					int x = (int) (Long.parseLong(st.nextToken()) / 10);
					int y = (int) (Long.parseLong(st.nextToken()) / 10);

					// 領域読み込み
					st = new StringTokenizer(bi.readLine(), ",");
					int width = Integer.parseInt(st.nextToken()) / 10;
					int height = Integer.parseInt(st.nextToken()) / 10;
					rect = new Rectangle(x, y, width, height);
				} finally {
					if (bi != null) {
						bi.close();
					}
				}
			}
		} else {
			rect = this.storage.readSLMStreaming(code);
		}
		return rect;
	}

	/**
	 * SLPファイルの読み込み
	 * 
	 * @param file
	 * @param area
	 * @return Point配列
	 * @throws IOException
	 * @throws NumberFormatException
	 */
	private Point[] readSLP(File file, Rectangle area)
			throws NumberFormatException, IOException {
		Map<Point, Point> map = new HashMap<Point, Point>();
		List<Point> slp = new ArrayList<Point>(this.DEFAULT_SLP_CAPACITY);
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new FileReader(file));
			while (bi.ready()) {
				String line = bi.readLine();
				int x = Integer.parseInt(line.substring(0, 7)) + area.x;
				int y = Integer.parseInt(line.substring(8, 15)) + area.y;
				Point p = new Point(x, y);
				if (map.containsKey(p)) {
					p = map.get(p);
				}
				slp.add(p);
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return slp.toArray(new Point[] {});
	}

	/**
	 * 駅区間の読み込み
	 * 
	 * @param file
	 * @param railway
	 * @return 駅区間の配列
	 * @throws IOException
	 */
	public Station[] readStation(File file, Map<Integer, Curve> railway)
			throws IOException {
		List<Station> station = new ArrayList<Station>();
		BufferedReader bi = null;
		try {
			bi = new BufferedReader(new InputStreamReader(new FileInputStream(
					file), "SJIS"));
			while (bi.ready()) {
				String line = bi.readLine();
				Matcher matcher = this.SAL_PARSER.matcher(line.substring(17));
				String name = null;
				Curve curve = null;
				while (matcher.find()) {
					String id = matcher.group(1);
					if (id.equals("NM")) {
						// 駅名
						name = matcher.group(3);
					} else if (id.equals("KN")) {
						curve = railway.get(Integer.valueOf(matcher.group(2)
								.substring(10, 16)));
					}
				}
				station.add(new Station(name, curve));
			}
		} finally {
			if (bi != null) {
				bi.close();
			}
		}
		return station.toArray(new Station[] {});
	}
}
